from bpeasy import train_bpe


def test_train_bpe_vocab_size():
    vocab_size = 300
    max_token_length = 4
    regex = r"([^\s]+)|(\s+)"
    vocab = train_bpe(
        iter(["This is a test", "this is another test", "good tests"]),
        regex,
        max_token_length,
        vocab_size,
    )
    assert len(vocab) == 267


def test_train_bpe_max_token_length():
    vocab_size = 300
    max_token_length = 2
    regex = r"([^\s]+)|(\s+)"
    vocab = train_bpe(
        iter(["This is a test", "this is another test", "good tests"]),
        regex,
        max_token_length,
        vocab_size,
    )
    for token in vocab:
        assert len(token) <= max_token_length
    max_token_length = 3
    vocab = train_bpe(
        iter(["This is a test", "this is another test", "good tests"]),
        regex,
        max_token_length,
        vocab_size,
    )
    for token in vocab:
        assert len(token) <= max_token_length


def test_train_bpe_gpt_regex():
    vocab_size = 300
    max_token_length = 128
    regex = r"""(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"""
    vocab = train_bpe(
        iter(["We've got a test", "We've got good test", "this is a good tests"]),
        regex,
        max_token_length,
        vocab_size,
    )
    for token in vocab:
        assert len(token) <= max_token_length

    print(vocab)
    assert b" go" in vocab.keys()
    assert b"'ve" in vocab.keys()



import json
import base64
from typing import Iterator

import tiktoken

from .bpeasy import train_bpe
from .convert import convert_tiktoken_to_huggingface


_DEFAULT_REGEX_PATTERN = r"""[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"""


class BPEasyTokenizer:
    def __init__(
        self,
        vocab: dict[bytes, int],
        regex_pattern: str = _DEFAULT_REGEX_PATTERN,
        special_tokens: list[str] = [],
        fill_to_nearest_multiple_of_eight=False,
        name="bpeasy",
    ):
        """
        Wrapper around tiktoken.Encoding
        Handles the loading/saving of vocab/special_tokens/regex
        """

        self.name = name
        self.regex_pattern = regex_pattern
        self.special_tokens = special_tokens
        self.vocab = vocab

        # Sort the vocab by rank
        sorted_vocab = sorted(list(vocab.items()), key=lambda x: x[1])

        # add special tokens
        special_token_ranks = {}
        for special_token in special_tokens:
            special_token_ranks[special_token] = len(sorted_vocab)
            sorted_vocab.append((special_token.encode("utf-8"), len(sorted_vocab)))

        full_vocab = dict(sorted_vocab)

        # fill to nearest multiple of 8
        if fill_to_nearest_multiple_of_eight:
            while len(sorted_vocab) % 8 != 0:
                sorted_vocab.append(
                    (
                        f"<|special-{len(sorted_vocab)}|>".encode("utf-8"),
                        len(sorted_vocab),
                    )
                )

        self._encoder = tiktoken.Encoding(
            name=name,
            pat_str=self.regex_pattern,
            mergeable_ranks=full_vocab,
            special_tokens=special_token_ranks,
        )

    def encode(self, text: str, **kwargs) -> list[int]:
        return self._encoder.encode(text, **kwargs)

    def decode(self, tokens: list[int], **kwargs) -> str:
        return self._encoder.decode(tokens, **kwargs)

    @classmethod
    def from_file(cls, file_path: str) -> "BPEasyTokenizer":
        with open(file_path, "r") as file:
            data = json.load(file)
            bytes_vocab = {
                base64.b64decode(key): value for key, value in data["vocab"].items()
            }
            instance = cls(
                name=data["name"],
                vocab=bytes_vocab,
                regex_pattern=data["regex_pattern"],
                special_tokens=data["special_tokens"],
            )
            return instance

    def save(self, file_path: str) -> None:
        with open(file_path, "w") as file:
            json.dump(
                {
                    "name": self.name,
                    "regex_pattern": self.regex_pattern,
                    "special_tokens": self.special_tokens,
                    "vocab": {
                        base64.b64encode(key).decode("utf-8"): value
                        for key, value in self.vocab.items()
                    },
                },
                file,
            )

    def export_to_huggingface_format(self, out_path: str) -> None:
        convert_tiktoken_to_huggingface(self._encoder, out_path, self.regex_pattern)

    def __len__(self) -> int:
        return len(self.vocab)

    @classmethod
    def train(
        cls,
        iterator: Iterator[str],
        vocab_size: int = 32_000,
        max_token_length=128,
        regex_pattern: str = _DEFAULT_REGEX_PATTERN,
        special_tokens: list[str] = [],
        fill_to_nearest_multiple_of_eight=False,
        name="bpeasy",
    ) -> "BPEasyTokenizer":
        bytes_vocab = train_bpe(iterator, regex_pattern, max_token_length, vocab_size)
        return cls(
            name=name,
            vocab=bytes_vocab,
            regex_pattern=regex_pattern,
            special_tokens=special_tokens,
            fill_to_nearest_multiple_of_eight=fill_to_nearest_multiple_of_eight,
        )



import base64
from unittest.mock import mock_open, patch, call

from bpeasy import save_vocab_to_tiktoken


def test_basic_functionality():
    vocab = {b"hello": 1, b"world": 2}
    with patch("builtins.open", mock_open()) as mock_file:
        save_vocab_to_tiktoken(vocab, "path/to/file.txt")
        mock_file.assert_called_once_with("path/to/file.txt", "wb")
        # Check if the sorted vocab is written correctly
        expected_content = [
            base64.b64encode(b"hello") + b" 1\n",
            base64.b64encode(b"world") + b" 2\n",
        ]
        mock_file().write.assert_has_calls(
            [call(content) for content in expected_content]
        )


def test_special_tokens_addition():
    vocab = {b"token": 0}
    special_tokens = ["special1", "special2"]
    with patch("builtins.open", mock_open()) as mock_file:
        save_vocab_to_tiktoken(vocab, "path/to/file.txt", special_tokens)
        # Check if special tokens are added correctly
        expected_content = [
            base64.b64encode(b"token") + b" 0\n",
            base64.b64encode("special1".encode("utf-8")) + b" 1\n",
            base64.b64encode("special2".encode("utf-8")) + b" 2\n",
        ]
        mock_file().write.assert_has_calls(
            [call(content) for content in expected_content]
        )


def test_fill_to_nearest_multiple_of_eight():
    vocab = {b"token": 0}
    with patch("builtins.open", mock_open()) as mock_file:
        save_vocab_to_tiktoken(
            vocab, "path/to/file.txt", fill_to_nearest_multiple_of_eight=True
        )
        # Verify that additional tokens are added to make the count a multiple of eight
        mock_file().write.assert_called()
        # Check the exact content based on your logic of filling to nearest multiple of eight
        assert mock_file().write.call_count == 8
